# -*- coding: utf-8 -*-
# © 2017 Creu Blanca
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).


def dict_compare(d1, d2):
    assert len(d1) == len(d2)
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    assert len(intersect_keys) == len(d1)
    for key in d1_keys:
        assert d1[key] == d2[key]


def rdns_to_map(data):
    return {x.split('=')[0]: x.split('=')[1] for x in data.split(',') if x}



from lxml import etree

from pgxmlsig.constants import DSigNs
from pgxmlsig.utils import create_node
from .constants import NS_MAP, EtsiNS


def create_qualifying_properties(node, name=None, etsi='xades'):
    obj_node = create_node('Object', node, DSigNs)
    qualifying = etree.SubElement(
        obj_node,
        etree.QName(EtsiNS, 'QualifyingProperties'),
        nsmap={etsi: EtsiNS}
    )
    qualifying.set('Target', '#' + node.get('Id'))
    if name is not None:
        qualifying.set('Id', name)
    return qualifying


def create_signed_properties(node, name=None, datetime=None):
    properties = create_node('SignedProperties', node, EtsiNS)
    if name is not None:
        properties.set('Id', name)
    signature_properties = create_node(
        'SignedSignatureProperties', properties, EtsiNS
    )
    signing_time = create_node('SigningTime', signature_properties, EtsiNS)
    if datetime is not None:
        signing_time.text = datetime.isoformat()
    create_node('SigningCertificate', signature_properties, EtsiNS)
    create_node('SignaturePolicyIdentifier', signature_properties, EtsiNS)
    return properties


def add_production_place(
        node, city=None, state=None, postal_code=None, country=None):
    signature_properties = node.find(
        'xades:SignedSignatureProperties', namespaces=NS_MAP
    )
    production_place = signature_properties.find(
        'xades:SignatureProductionPlace', namespaces=NS_MAP
    )
    if production_place is None:
        production_place = create_node(
            'SignatureProductionPlace', ns=EtsiNS
        )
        signature_properties.insert(3, production_place)
    for child in production_place.getchildren():
        production_place.remove(child)
    if city is not None:
        create_node('City', production_place, EtsiNS).text = city
    if state is not None:
        create_node('StateOrProvince', production_place, EtsiNS).text = state
    if postal_code is not None:
        create_node('PostalCode', production_place, EtsiNS).text = postal_code
    if country is not None:
        create_node('CountryName', production_place, EtsiNS).text = country


def add_claimed_role(node, role):
    signature_properties = node.find(
        'xades:SignedSignatureProperties', namespaces=NS_MAP
    )
    signer_role = signature_properties.find(
        'xades:SignerRole', namespaces=NS_MAP
    )
    if signer_role is None:
        signer_role = create_node(
            'SignerRole', signature_properties, ns=EtsiNS
        )
    claimed_roles = signer_role.find(
        'xades:ClaimedRoles', namespaces=NS_MAP
    )
    if claimed_roles is None:
        claimed_roles = create_node(
            'ClaimedRoles', ns=EtsiNS
        )
        signer_role.insert(0, claimed_roles)
    claimed_role = create_node('ClaimedRole', claimed_roles, EtsiNS)
    claimed_role.text = role
    return claimed_role



# -*- coding: utf-8 -*-
# © 2017 Creu Blanca
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).

from cryptography.hazmat.primitives import hashes

from pgxmlsig import constants
from .ns import EtsiNS



NS_MAP = {
    'ds': 'http://www.w3.org/2000/09/xmldsig#',
    'fe' : 'http://www.dian.gov.co/contratos/facturaelectronica/v1',
    'cac' : 'urn:oasis:names:specification:ubl:schema:xsd:CommonAggregateComponents-2',
    'cbc' : 'urn:oasis:names:specification:ubl:schema:xsd:CommonBasicComponents-2',
    'clm54217' : 'urn:un:unece:uncefact:codelist:specification:54217:2001',
    'clm66411' : 'urn:un:unece:uncefact:codelist:specification:66411:2001',
    'clmIANAMIMEMediaType' : 'urn:un:unece:uncefact:codelist:specification:IANAMIMEMediaType:2003',
    'ext' : 'urn:oasis:names:specification:ubl:schema:xsd:CommonExtensionComponents-2',
    'qdt' : 'urn:oasis:names:specification:ubl:schema:xsd:QualifiedDatatypes-2',
    'sts' : 'http://www.dian.gov.co/contratos/facturaelectronica/v1/Structures',
    'udt' : 'urn:un:unece:uncefact:data:specification:UnqualifiedDataTypesSchemaModule:2',
    'xsi' : 'http://www.w3.org/2001/XMLSchema-instance',
    'ds11': 'http://www.w3.org/2009/xmldsig11#',
    'xades': 'http://uri.etsi.org/01903/v1.3.2#',
    'xades141': 'http://uri.etsi.org/01903/v1.4.1#'
    }

MAP_HASHLIB = {
    constants.TransformMd5: hashes.MD5,
    constants.TransformSha1: hashes.SHA1,
    constants.TransformSha224: hashes.SHA224,
    constants.TransformSha256: hashes.SHA256,
    constants.TransformSha384: hashes.SHA384,
    constants.TransformSha512: hashes.SHA512,
}



import sys

from setuptools import find_packages, setup

install_requires = [
    'pytz',
    'pg-xmlsig'
]

setup(
    name='pg-xades',
    version='0.0.3',
    description='XaDES XML signature. Based on python-xades.',
    long_description='XaDES XML Signature created with cryptography and lxml',
    author="Jaime Bermeo",
    author_email="jaime.bermeo@pragmatic.com.co",
    url='https://github.com/Silverdoses/python-pgxades',

    install_requires=install_requires,
    entry_points={},
    package_dir={'': 'src'},
    packages=find_packages('src'),
    include_package_data=True,

    license='BSD',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# © 2017 Creu Blanca
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html).
# Modificado por Pragmatic SAS para ajustar a los requerimientos de Colombia

from base64 import b64decode, b64encode
import hashlib


from .ns import EtsiNS
from pgxades.constants import NS_MAP, MAP_HASHLIB
from cryptography.hazmat.backends import default_backend
from cryptography.x509 import load_der_x509_certificate
from pgxades.utils import rdns_to_map, dict_compare
from pgxmlsig.constants import TransformUsageDigestMethod
from pgxmlsig.ns import DSigNs
from pgxmlsig.utils import create_node, USING_PYTHON2, get_rdns_name

if USING_PYTHON2:
    import urllib
else:
    import urllib.request as urllib







class Policy(object):
    """"
    Policy class created in order to define different policies
    """

    hash_method = None

    def sign(self, signature):
        return

    def validate(self, signature):
        return

    def calculate_policy_node(self, node, sign=False):
        if sign:
            return create_node('SignaturePolicyImplied', node, EtsiNS)
        return node.find('xades:SignaturePolicyImplied', namespaces=NS_MAP)

    def calculate_certificates(self, node, key_x509):
        self.calculate_certificate(node, key_x509)

    def calculate_certificate(self, node, key_x509):
        cert = create_node('Cert', node, EtsiNS)
        cert_digest = create_node('CertDigest', cert, EtsiNS)
        digest_algorithm = create_node('DigestMethod', cert_digest, DSigNs)
        digest_algorithm.set('Algorithm', self.hash_method)
        digest_value = create_node('DigestValue', cert_digest, DSigNs)
        digest_value.text = b64encode(key_x509.fingerprint(
            MAP_HASHLIB[self.hash_method]()
        ))
        issuer_serial = create_node('IssuerSerial', cert, EtsiNS)
        create_node(
            'X509IssuerName', issuer_serial, DSigNs
        ).text = get_rdns_name(key_x509.issuer.rdns)
        create_node(
            'X509SerialNumber', issuer_serial, DSigNs
        ).text = str(key_x509.serial_number)
        return

    def validate_certificate(self, node, signature):
        certs = node.findall('xades:Cert', namespaces=NS_MAP)
        x509 = signature.find('ds:KeyInfo/ds:X509Data', namespaces=NS_MAP)
        x509_data = x509.find('ds:X509Certificate', namespaces=NS_MAP)
        serial = x509.find('ds:X509IssuerSerial', namespaces=NS_MAP)
        if serial is not None:
            serial_name = serial.find(
                'ds:X509IssuerName', namespaces=NS_MAP
            ).text
            serial_number = serial.find(
                'ds:X509SerialNumber', namespaces=NS_MAP
            ).text
            certificate = None
            for cert in certs:
                print(cert.find(
                        'xades:IssuerSerial/ds:X509IssuerName',
                        namespaces=NS_MAP
                ).text)
                print(serial_name)
                if cert.find(
                        'xades:IssuerSerial/ds:X509IssuerName',
                        namespaces=NS_MAP
                ).text == serial_name and cert.find(
                    'xades:IssuerSerial/ds:X509SerialNumber',
                    namespaces=NS_MAP
                ).text == serial_number:
                    certificate = cert

            assert certificate is not None
        else:
            certificate = certs[0]
        if x509_data is not None:
            parsed_x509 = load_der_x509_certificate(
                b64decode(x509_data.text), default_backend()
            )
            assert str(parsed_x509.serial_number) == certificate.find(
                'xades:IssuerSerial/ds:X509SerialNumber', namespaces=NS_MAP
            ).text
            dict_compare(
                rdns_to_map(get_rdns_name(parsed_x509.issuer.rdns)),
                rdns_to_map(certificate.find(
                    'xades:IssuerSerial/ds:X509IssuerName',
                    namespaces=NS_MAP
                ).text)
            )
            digest = certificate.find(
                'xades:CertDigest', namespaces=NS_MAP
            )
            assert b64encode(
                parsed_x509.fingerprint(MAP_HASHLIB[digest.find(
                    'ds:DigestMethod', namespaces=NS_MAP
                ).get('Algorithm')]())) == digest.find(
                'ds:DigestValue', namespaces=NS_MAP).text.encode()





class PolicyId(Policy):
    id = None
    name = None
    remote = None

    def set_transforms(self, node, value, sign=False):
        """
        Transformations of the policy if required. Modifies node and returns
        transformed value
        :param node: Policy node
        :param value: Original value
        :return: str
        """
        return value

    def calculate_policy_node(self, node, sign=False):
        if sign:
            policy_id = create_node('SignaturePolicyId', node, EtsiNS)
            identifier = create_node('SigPolicyId', policy_id, EtsiNS)
            create_node('Identifier', identifier, EtsiNS).text = self.id
            create_node('Description', identifier, EtsiNS).text = self.name
            remote = self.remote or self.id
        else:
            policy_id = node.find('xades:SignaturePolicyId', namespaces=NS_MAP)
            identifier = policy_id.find('xades:SigPolicyId',  namespaces=NS_MAP)
            remote = identifier.find('xades:Identifier', namespaces=NS_MAP).text

        if self.remote:
            value = self.remote  # Base64-decoded PDF (bytes)
        else:
            value = urllib.urlopen(remote).read()
        value = self.set_transforms(policy_id, value, sign)
        if sign:
            hash_method = self.hash_method
            digest = create_node('SigPolicyHash', policy_id, EtsiNS)
            digest_method = create_node('DigestMethod', digest, DSigNs)
            digest_method.set('Algorithm', self.hash_method)
            digest_value = create_node('DigestValue', digest, DSigNs)
        else:
            hash_method = policy_id.find(
                'xades:SigPolicyHash/ds:DigestMethod', namespaces=NS_MAP
            ).get('Algorithm')
            digest_value  = policy_id.find(
                'xades:SigPolicyHash/ds:DigestValue', namespaces=NS_MAP
            )
        hash_calc = hashlib.new(TransformUsageDigestMethod[hash_method])
        hash_calc.update(value)
        digest_val = hash_calc.digest()
        if sign:
            digest_value.text = b64encode(digest_val)
        assert digest_value.text.encode() == b64encode(digest_val)
        return policy_id



